@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
  *
- * Copyright (C) 2004-2005 William Jon McCann <mccann@jhu.edu>
+ * Copyright (C) 2004-2006 William Jon McCann <mccann@jhu.edu>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -24,9 +24,11 @@
 
 #include "config.h"
 
+#define _GNU_SOURCE
 #include <unistd.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/types.h>		/* For uid_t, gid_t */
 
 #include <glib/gi18n.h>
 #include <gdk/gdkx.h>
@@ -42,7 +44,7 @@
 
 #include "file-transfer-dialog.h"
 
-#include "gs-visual-gl.h"
+#include "gs-theme-manager.h"
 #include "gs-job.h"
 #include "gs-prefs.h" /* for GS_MODE enum */
 
@@ -82,8 +84,9 @@ static GtkTargetEntry drop_types [] =
         { "_NETSCAPE_URL", 0, TARGET_NS_URL }
 };
 
-static GladeXML *xml = NULL;
-static GSJob    *job = NULL;
+static GladeXML       *xml = NULL;
+static GSThemeManager *theme_manager = NULL;
+static GSJob          *job = NULL;
 
 static gint32
 config_get_activate_delay (gboolean *is_writable)
@@ -101,8 +104,9 @@ config_get_activate_delay (gboolean *is_writable)
 
         delay = gconf_client_get_int (client, KEY_ACTIVATE_DELAY, NULL);
 
-        if (delay < 1)
+        if (delay < 1) {
                 delay = 1;
+        }
 
         g_object_unref (client);
 
@@ -213,18 +217,18 @@ config_get_theme (gboolean *is_writable)
 }
 
 static GSList *
-get_all_theme_ids (GSJob *job)
+get_all_theme_ids (GSThemeManager *theme_manager)
 {
         GSList *ids = NULL;
         GSList *entries;
         GSList *l;
 
-        entries = gs_job_get_theme_info_list (job);
+        entries = gs_theme_manager_get_info_list (theme_manager);
         for (l = entries; l; l = l->next) {
-                GSJobThemeInfo *info = l->data;
+                GSThemeInfo *info = l->data;
 
-                ids = g_slist_prepend (ids, g_strdup (gs_job_theme_info_get_id (info)));
-                gs_job_theme_info_unref (info);
+                ids = g_slist_prepend (ids, g_strdup (gs_theme_info_get_id (info)));
+                gs_theme_info_unref (info);
         }
         g_slist_free (entries);
 
@@ -246,7 +250,7 @@ config_set_theme (const char *theme_id)
                 mode = GS_MODE_RANDOM;
 
                 /* set the themes key to contain all available screensavers */
-                list = get_all_theme_ids (job);
+                list = get_all_theme_ids (theme_manager);
         } else {
                 mode = GS_MODE_SINGLE;
                 list = g_slist_append (list, g_strdup (theme_id));
@@ -341,16 +345,28 @@ preview_clear (GtkWidget *widget)
         gdk_window_clear (widget->window);
 }
 
+static void
+job_set_theme (GSJob      *job,
+               const char *theme)
+{
+        GSThemeInfo *info;
+        const char  *command;
+
+        info = gs_theme_manager_lookup_theme_info (theme_manager, theme);
+        command = gs_theme_info_get_exec (info);
+        gs_job_set_command (job, command);
+        gs_theme_info_unref (info);
+}
+
 static void
 preview_set_theme (GtkWidget  *widget,
                    const char *theme,
                    const char *name)
 {
-        GError    *error = NULL;
         GtkWidget *label;
         char      *markup;
 
-        if (job) {
+        if (job != NULL) {
                 gs_job_stop (job);
         }
 
@@ -366,29 +382,16 @@ preview_set_theme (GtkWidget  *widget,
         } else if (theme && strcmp (theme, "__random") == 0) {
                 GSList *themes;
 
-                themes = get_all_theme_ids (job);
-                if (themes) {
-                        if (! gs_job_set_theme (job, (const char *) themes->data, &error)) {
-                                if (error) {
-                                        g_warning ("Could not set theme: %s", error->message);
-                                        g_error_free (error);
-                                }
-                                return;
-                        }
+                themes = get_all_theme_ids (theme_manager);
+                if (themes != NULL) {
+                        job_set_theme (job, (const char *) themes->data);
                         g_slist_foreach (themes, (GFunc) g_free, NULL);
                         g_slist_free (themes);
 
                         gs_job_start (job);
                 }
-                
         } else {
-                if (! gs_job_set_theme (job, theme, &error)) {
-                        if (error) {
-                                g_warning ("Could not set theme: %s", error->message);
-                                g_error_free (error);
-                        }
-                        return;
-                }
+                job_set_theme (job, theme);
                 gs_job_start (job);
         }
 }
@@ -421,7 +424,7 @@ response_cb (GtkWidget *widget,
 static GSList *
 get_theme_info_list (void)
 {
-        return gs_job_get_theme_info_list (job);
+        return gs_theme_manager_get_info_list (theme_manager);
 }
 
 static void
@@ -451,22 +454,23 @@ populate_model (GtkTreeStore *store)
 
         themes = get_theme_info_list ();
 
-        if (! themes)
+        if (themes == NULL) {
                 return;
-        
+        }
+
         for (l = themes; l; l = l->next) {
-                GSJobThemeInfo *info = l->data;
+                GSThemeInfo *info = l->data;
 
                 if (! info)
                         continue;
 
                 gtk_tree_store_append (store, &iter, NULL);
                 gtk_tree_store_set (store, &iter,
-                                    NAME_COLUMN, gs_job_theme_info_get_name (info),
-                                    ID_COLUMN, gs_job_theme_info_get_id (info),
+                                    NAME_COLUMN, gs_theme_info_get_name (info),
+                                    ID_COLUMN, gs_theme_info_get_id (info),
                                     -1);
 
-                gs_job_theme_info_unref (info);
+                gs_theme_info_unref (info);
         }
 
         g_slist_free (themes);
@@ -520,7 +524,7 @@ tree_selection_changed_cb (GtkTreeSelection *selection,
 
         gtk_tree_model_get (model, &iter, ID_COLUMN, &theme, NAME_COLUMN, &name, -1);
 
-        if (! theme) {
+        if (theme == NULL) {
                 g_free (name);
                 return;
         }
@@ -612,12 +616,13 @@ separator_func (GtkTreeModel *model,
 {
         int   column = GPOINTER_TO_INT (data);
         char *text;
-        
+
         gtk_tree_model_get (model, iter, column, &text, -1);
-        
-        if (text && strcmp (text, "__separator") == 0)
+
+        if (text != NULL && strcmp (text, "__separator") == 0) {
                 return TRUE;
-        
+        }
+
         g_free (text);
 
         return FALSE;
@@ -761,26 +766,38 @@ theme_installer_run (GtkWidget *parent,
 {
         GtkWidget    *dialog;
         GnomeVFSURI  *src_uri;
-        GList        *src, *target;
-        char         *target_path;
+        GList        *src;
+        GList        *target;
         char         *user_dir;
-        char         *base;
+        char         *short_name;
+        char         *base_name;
+        char         *target_path;
+        gboolean      is_desktop;
 
         src_uri = gnome_vfs_uri_new (filename);
         src = g_list_append (NULL, src_uri);
-
-        user_dir = g_build_filename (g_get_user_data_dir (), "gnome-screensaver", "themes", NULL);
-        base = gnome_vfs_uri_extract_short_name (src_uri);
-
+        target = NULL;
         target_path = NULL;
 
-        while (TRUE) {
-                char      *file_tmp;
-                int        len = strlen (base);
+        user_dir = g_build_filename (g_get_user_data_dir (), "applications", "screensavers", NULL);
+        g_mkdir_with_parents (user_dir, S_IRWXU);
 
-                if (base && len > 4 && (!strcmp (base + len - 4, ".xml"))) {
-                        file_tmp = g_strdup_printf ("screensaver-theme-%d.xml", rand ());
-                } else {
+        short_name = gnome_vfs_uri_extract_short_name (src_uri);
+        base_name = NULL;
+
+        is_desktop = FALSE;
+        if (short_name != NULL
+            && g_str_has_suffix (short_name, ".desktop")) {
+                /* FIXME: validate key file? */
+                is_desktop = TRUE;
+                base_name = g_strndup (short_name, strlen (short_name) - 8);
+        }
+
+        while (TRUE) {
+                char *tmp;
+                gboolean exists;
+
+                if (! is_desktop) {
                         dialog = gtk_message_dialog_new (GTK_WINDOW (parent),
                                                          GTK_DIALOG_MODAL,
                                                          GTK_MESSAGE_ERROR,
@@ -795,19 +812,24 @@ theme_installer_run (GtkWidget *parent,
                         gtk_dialog_run (GTK_DIALOG (dialog));
                         gtk_widget_destroy (dialog);
                         g_free (target_path);
-                        return;
+                        goto out;
                 }
 
-                target_path = g_build_filename (user_dir, file_tmp, NULL);
+                g_free (target_path);
+                target_path = g_build_filename (user_dir, short_name, NULL);
 
-                g_free (file_tmp);
-                if (! gnome_vfs_uri_exists (gnome_vfs_uri_new (target_path)))
+                /* FIXME: racy, oh well */
+                exists = gnome_vfs_uri_exists (gnome_vfs_uri_new (target_path));
+                if (! exists) {
+                        target = g_list_append (NULL, gnome_vfs_uri_new (target_path));
                         break;
-        }
-                
-        g_free (user_dir);
+                }
 
-        target = g_list_append (NULL, gnome_vfs_uri_new (target_path));
+                /* try another name */
+                tmp = g_strdup_printf ("%s-%u.desktop",  base_name, g_random_int ());
+                g_free (short_name);
+                short_name = tmp;
+        }
 
         dialog = file_transfer_dialog_new ();
         gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (parent));
@@ -819,8 +841,6 @@ theme_installer_run (GtkWidget *parent,
                                               GNOME_VFS_XFER_ERROR_MODE_QUERY,
                                               GNOME_VFS_XFER_OVERWRITE_MODE_QUERY,
                                               GNOME_VFS_PRIORITY_DEFAULT);
-        gnome_vfs_uri_list_unref (src);
-        gnome_vfs_uri_list_unref (target);
 
         g_signal_connect (dialog, "cancel",
                           G_CALLBACK (transfer_cancel_cb), target_path);
@@ -828,6 +848,14 @@ theme_installer_run (GtkWidget *parent,
                           G_CALLBACK (transfer_done_cb), target_path);
 
         gtk_widget_show (dialog);
+
+ out:
+        gnome_vfs_uri_list_unref (src);
+        gnome_vfs_uri_list_unref (target);
+
+        g_free (user_dir);
+        g_free (base_name);
+        g_free (short_name);
 }
 
 /* Callback issued during drag movements */
@@ -874,14 +902,13 @@ drag_data_received_cb (GtkWidget        *widget,
         if (uris != NULL && uris->data != NULL) {
                 GnomeVFSURI *uri = (GnomeVFSURI *) uris->data;
 
-                if (gnome_vfs_uri_is_local (uri))
-                        filename = gnome_vfs_unescape_string (
-                                        gnome_vfs_uri_get_path (uri),
-                                        G_DIR_SEPARATOR_S);
-                else
-                        filename = gnome_vfs_unescape_string (
-                                        gnome_vfs_uri_to_string (uri, GNOME_VFS_URI_HIDE_NONE),
-                                        G_DIR_SEPARATOR_S);
+                if (gnome_vfs_uri_is_local (uri)) {
+                        filename = gnome_vfs_unescape_string (gnome_vfs_uri_get_path (uri),
+                                                              G_DIR_SEPARATOR_S);
+                } else {
+                        filename = gnome_vfs_unescape_string (gnome_vfs_uri_to_string (uri, GNOME_VFS_URI_HIDE_NONE),
+                                                              G_DIR_SEPARATOR_S);
+                }
 
                 gnome_vfs_uri_list_unref (uris);
         }
@@ -1081,7 +1108,7 @@ key_changed_cb (GConfClient *client,
 
 static void
 fullscreen_preview_previous_cb (GtkWidget *fullscreen_preview_window,
-                                gpointer   user_data) 
+                                gpointer   user_data)
 {
         GtkWidget        *treeview;
         GtkTreeSelection *selection;
@@ -1093,7 +1120,7 @@ fullscreen_preview_previous_cb (GtkWidget *fullscreen_preview_window,
 
 static void
 fullscreen_preview_next_cb (GtkWidget *fullscreen_preview_window,
-                            gpointer   user_data) 
+                            gpointer   user_data)
 {
         GtkWidget        *treeview;
         GtkTreeSelection *selection;
@@ -1105,7 +1132,7 @@ fullscreen_preview_next_cb (GtkWidget *fullscreen_preview_window,
 
 static void
 fullscreen_preview_cancelled_cb (GtkWidget *button,
-                                 gpointer   user_data) 
+                                 gpointer   user_data)
 {
 
         GtkWidget *fullscreen_preview_area;
@@ -1118,7 +1145,7 @@ fullscreen_preview_cancelled_cb (GtkWidget *button,
 
         fullscreen_preview_area = glade_xml_get_widget (xml, "fullscreen_preview_area");
         preview_clear (fullscreen_preview_area);
-        
+
         fullscreen_preview_window = glade_xml_get_widget (xml, "fullscreen_preview_window");
         gtk_widget_hide (fullscreen_preview_window);
 
@@ -1137,14 +1164,14 @@ fullscreen_preview_start_cb (GtkWidget *widget,
 
         dialog = glade_xml_get_widget (xml, "prefs_dialog");
         gtk_widget_hide (dialog);
-        
+
         fullscreen_preview_window = glade_xml_get_widget (xml, "fullscreen_preview_window");
         gtk_widget_show (fullscreen_preview_window);
         gtk_widget_grab_focus (fullscreen_preview_window);
 
         gtk_window_fullscreen (GTK_WINDOW (fullscreen_preview_window));
         gtk_window_set_keep_above (GTK_WINDOW (fullscreen_preview_window), TRUE);
-				
+
         fullscreen_preview_area = glade_xml_get_widget (xml, "fullscreen_preview_area");
         gs_job_set_widget (job, fullscreen_preview_area);
 }
@@ -1173,6 +1200,133 @@ setup_list_size_constraint (GtkWidget *widget,
                           G_CALLBACK (constrain_list_size), to_size);
 }
 
+static gboolean
+check_is_root_user (void)
+{
+#ifndef G_OS_WIN32
+  uid_t ruid, euid, suid; /* Real, effective and saved user ID's */
+  gid_t rgid, egid, sgid; /* Real, effective and saved group ID's */
+
+#ifdef HAVE_GETRESUID
+  if (getresuid (&ruid, &euid, &suid) != 0 ||
+      getresgid (&rgid, &egid, &sgid) != 0)
+#endif /* HAVE_GETRESUID */
+    {
+      suid = ruid = getuid ();
+      sgid = rgid = getgid ();
+      euid = geteuid ();
+      egid = getegid ();
+    }
+
+  if (ruid == 0) {
+          return TRUE;
+  }
+
+#endif
+  return FALSE;
+}
+
+static void
+setup_for_root_user (void)
+{
+        GtkWidget *lock_checkbox;
+        GtkWidget *label;
+
+        lock_checkbox = glade_xml_get_widget (xml, "lock_checkbox");
+        label = glade_xml_get_widget (xml, "root_warning_label");
+
+        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (lock_checkbox), FALSE);
+        gtk_widget_set_sensitive (lock_checkbox, FALSE);
+
+        gtk_widget_show (label);
+}
+
+static GdkVisual *
+get_best_visual (void)
+{
+        char         *command;
+        char         *std_output;
+        int           exit_status;
+        GError       *error;
+        unsigned long v;
+        char          c;
+        GdkVisual    *visual;
+
+        visual = NULL;
+
+        command = g_build_filename (LIBEXECDIR, "gnome-screensaver-gl-helper", NULL);
+
+        error = NULL;
+        g_spawn_command_line_sync (command,
+                                   &std_output,
+                                   NULL,
+                                   &exit_status,
+                                   &error);
+
+        if (1 == sscanf (std_output, "0x%lx %c", &v, &c)) {
+                if (v != 0) {
+                        VisualID      visual_id;
+
+                        visual_id = (VisualID) v;
+                        visual = gdkx_visual_get (visual_id);
+
+                        g_message ("Found best visual for GL: 0x%x",
+                                   (unsigned int) visual_id);
+                }
+        }
+
+        g_free (std_output);
+
+        return visual;
+}
+
+static GdkColormap *
+get_best_colormap_for_screen (GdkScreen *screen)
+{
+        GdkColormap *colormap;
+        GdkVisual   *visual;
+
+        g_return_val_if_fail (screen != NULL, NULL);
+
+        visual = get_best_visual ();
+
+        colormap = NULL;
+        if (visual != NULL) {
+                colormap = gdk_colormap_new (visual, FALSE);
+        }
+
+        return colormap;
+}
+
+static void
+widget_set_best_colormap (GtkWidget *widget)
+{
+        GdkColormap *colormap;
+
+        g_return_if_fail (widget != NULL);
+
+        colormap = get_best_colormap_for_screen (gtk_widget_get_screen (widget));
+        if (colormap != NULL) {
+                gtk_widget_set_colormap (widget, colormap);
+                g_object_unref (colormap);
+        }
+}
+
+static gboolean
+setup_treeview_idle (gpointer data)
+{
+        GtkWidget *preview;
+        GtkWidget *treeview;
+
+        preview  = glade_xml_get_widget (xml, "preview_area");
+        treeview = glade_xml_get_widget (xml, "savers_treeview");
+
+        setup_treeview (treeview, preview);
+        setup_treeview_selection (treeview);
+
+        return FALSE;
+}
+
 static void
 init_capplet (void)
 {
@@ -1185,6 +1339,7 @@ init_capplet (void)
         GtkWidget *label;
         GtkWidget *enabled_checkbox;
         GtkWidget *lock_checkbox;
+        GtkWidget *root_warning_label;
         GtkWidget *preview_button;
         GtkWidget *fullscreen_preview_window;
         GtkWidget *fullscreen_preview_previous;
@@ -1226,6 +1381,7 @@ init_capplet (void)
         activate_delay_hbox   = glade_xml_get_widget (xml, "activate_delay_hbox");
         enabled_checkbox   = glade_xml_get_widget (xml, "enable_checkbox");
         lock_checkbox      = glade_xml_get_widget (xml, "lock_checkbox");
+        root_warning_label = glade_xml_get_widget (xml, "root_warning_label");
         preview_button     = glade_xml_get_widget (xml, "preview_button");
         fullscreen_preview_window = glade_xml_get_widget (xml, "fullscreen_preview_window");
         fullscreen_preview_area = glade_xml_get_widget (xml, "fullscreen_preview_area");
@@ -1238,25 +1394,29 @@ init_capplet (void)
         label              = glade_xml_get_widget (xml, "savers_label");
         gtk_label_set_mnemonic_widget (GTK_LABEL (label), treeview);
 
-        gs_visual_gl_widget_set_best_colormap (preview);
+        gtk_widget_set_no_show_all (root_warning_label, TRUE);
+        widget_set_best_colormap (preview);
 
         activate_delay = config_get_activate_delay (&is_writable);
         ui_set_delay (activate_delay);
-        if (! is_writable)
+        if (! is_writable) {
                 gtk_widget_set_sensitive (activate_delay_hbox, FALSE);
+        }
         g_signal_connect (activate_delay_hscale, "format-value",
                           G_CALLBACK (format_value_callback_time), NULL);
 
         gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (lock_checkbox), config_get_lock (&is_writable));
-        if (! is_writable)
+        if (! is_writable) {
                 gtk_widget_set_sensitive (lock_checkbox, FALSE);
+        }
         g_signal_connect (lock_checkbox, "toggled",
                           G_CALLBACK (lock_checkbox_toggled), NULL);
 
         enabled = config_get_enabled (&is_writable);
         ui_set_enabled (enabled);
-        if (! is_writable)
+        if (! is_writable) {
                 gtk_widget_set_sensitive (enabled_checkbox, FALSE);
+        }
         g_signal_connect (enabled_checkbox, "toggled",
                           G_CALLBACK (enabled_checkbox_toggled), NULL);
 
@@ -1281,7 +1441,7 @@ init_capplet (void)
         /* Update list of themes if using random screensaver */
         client = gconf_client_get_default ();
         string = gconf_client_get_string (client, KEY_MODE, NULL);
-        if (string) {
+        if (string != NULL) {
                 int mode;
                 GSList *list;
 
@@ -1289,7 +1449,7 @@ init_capplet (void)
                 g_free (string);
 
                 if (mode == GS_MODE_RANDOM) {
-                        list = get_all_theme_ids (job);
+                        list = get_all_theme_ids (theme_manager);
                         gconf_client_set_list (client, KEY_THEMES, GCONF_VALUE_STRING, list, NULL);
 
                         g_slist_foreach (list, (GFunc) g_free, NULL);
@@ -1310,8 +1470,9 @@ init_capplet (void)
         preview_clear (preview);
         gs_job_set_widget (job, preview);
 
-        setup_treeview (treeview, preview);
-        setup_treeview_selection (treeview);
+        if (check_is_root_user ()) {
+                setup_for_root_user ();
+        }
 
         g_signal_connect (activate_delay_hscale, "value-changed",
                           G_CALLBACK (activate_delay_value_changed_cb), NULL);
@@ -1329,6 +1490,8 @@ init_capplet (void)
                           G_CALLBACK (fullscreen_preview_previous_cb), NULL);
         g_signal_connect (fullscreen_preview_next, "clicked", 
                           G_CALLBACK (fullscreen_preview_next_cb), NULL);
+
+        g_idle_add ((GSourceFunc)setup_treeview_idle, NULL);
 }
 
 int
@@ -1340,9 +1503,9 @@ main (int    argc,
         bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
 # ifdef HAVE_BIND_TEXTDOMAIN_CODESET
         bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
-# endif 
+# endif
         textdomain (GETTEXT_PACKAGE);
-#endif 
+#endif
 
         gnome_program_init (PACKAGE, VERSION,
                             LIBGNOMEUI_MODULE,
@@ -1351,11 +1514,13 @@ main (int    argc,
                             NULL);
 
         job = gs_job_new ();
+        theme_manager = gs_theme_manager_new ();
 
         init_capplet ();
 
         gtk_main ();
 
+        g_object_unref (theme_manager);
         g_object_unref (job);
 
 	return 0;
